import 'dart:convert';

import 'package:dio/dio.dart';
import 'package:recase/recase.dart';

enum MorphNetworkErrorType {
  /// internal_server_error
  internalServerError,

  /// maintenance_error
  maintenanceError,

  /// request_error
  requestError,

  /// auth_error
  authError,

  /// user_error
  userError,

  /// team_error
  teamError,

  /// resource_error
  resourceError,

  /// secret_error
  secretError,

  /// application_error
  applicationError,

  /// table_error
  tableError,

  /// field_error
  fieldError,

  /// record_error
  recordError,

  /// userpool_user_error
  userpoolUserError,

  /// group_error
  groupError,
}

MorphNetworkError generateMorphNetworkError(DioError e) {
  switch (e.type) {
    case DioErrorType.cancel:
      return RequestCancelledError.fromDioError(e);
    case DioErrorType.connectTimeout:
      return RequestTimeoutError.fromDioError(e);
    case DioErrorType.sendTimeout:
      return SendTimeoutError.fromDioError(e);
    case DioErrorType.receiveTimeout:
      return ReceiveTimeoutError.fromDioError(e);
    case DioErrorType.response:
      switch (e.response?.statusCode) {
        case 400:
          if (e.response?.data is Map<String, dynamic>) {
            return generateMorphNetworkResponseError(
              e.requestOptions,
              e.response!.data as Map<String, dynamic>,
            );
          } else {
            return BadRequestError.fromDioError(e);
          }
        case 401:
        case 403:
          return UnauthorizedError.fromDioError(e);
        case 404:
          return NotFoundError.fromDioError(e);
        case 408:
          return RequestTimeoutError.fromDioError(e);
        case 409:
          return ConflictError.fromDioError(e);
        case 500:
          return InternalServerErrorError.fromDioError(e);
        case 503:
          return ServiceUnavailableError.fromDioError(e);
        default:
          return UnexpectedError.fromDioError(e);
      }
    case DioErrorType.other:
      return MorphNetworkError.fromDioError(e);
  }
}

MorphNetworkResponseError generateMorphNetworkResponseError(
  RequestOptions requestOptions,
  Map<String, dynamic> responseBody,
) {
  final responseError =
      MorphNetworkResponseError.fromResponseBody(requestOptions, responseBody);
  switch (responseError.errorType) {
    case MorphNetworkErrorType.requestError:
      // RequestError
      return BadRequestDetailResponseError.fromResponseError(responseError);
    case MorphNetworkErrorType.userpoolUserError:
      // UserpoolUserError
      switch (responseError.subCode) {
        case 1:
          return UserpoolUserNotFoundResponseError.fromResponseError(
            responseError,
          );
        default:
      }
      break;
    default:
  }
  return responseError;
}

/// Network base error
class MorphNetworkError extends DioError {
  MorphNetworkError({
    required super.requestOptions,
    super.response,
    super.type,
    super.error,
  });

  factory MorphNetworkError.fromDioError(DioError e) => MorphNetworkError(
        requestOptions: e.requestOptions,
        response: e.response,
        type: e.type,
        error: e.error,
      );
}

/// DioErrorType.cancel
class RequestCancelledError extends MorphNetworkError {
  RequestCancelledError({
    required super.requestOptions,
    super.response,
    super.type,
    super.error,
  });

  factory RequestCancelledError.fromDioError(DioError e) =>
      RequestCancelledError(
        requestOptions: e.requestOptions,
        response: e.response,
        type: e.type,
        error: e.error,
      );

  @override
  String toString() {
    return 'Request cancelled';
  }
}

/// DioErrorType.connectTimeout || (DioErrorType.response && statusCode == 404)
class RequestTimeoutError extends MorphNetworkError {
  RequestTimeoutError({
    required super.requestOptions,
    super.response,
    super.type,
    super.error,
  });

  factory RequestTimeoutError.fromDioError(DioError e) => RequestTimeoutError(
        requestOptions: e.requestOptions,
        response: e.response,
        type: e.type,
        error: e.error,
      );

  @override
  String toString() {
    return 'Invalid request';
  }
}

/// DioErrorType.sendTimeout
class SendTimeoutError extends MorphNetworkError {
  SendTimeoutError({
    required super.requestOptions,
    super.response,
    super.type,
    super.error,
  });

  factory SendTimeoutError.fromDioError(DioError e) => SendTimeoutError(
        requestOptions: e.requestOptions,
        response: e.response,
        type: e.type,
        error: e.error,
      );

  @override
  String toString() {
    return 'Send timeout in connection with API server, please try again';
  }
}

/// DioErrorType.receiveTimeout
class ReceiveTimeoutError extends MorphNetworkError {
  ReceiveTimeoutError({
    required super.requestOptions,
    super.response,
    super.type,
    super.error,
  });

  factory ReceiveTimeoutError.fromDioError(DioError e) => ReceiveTimeoutError(
        requestOptions: e.requestOptions,
        response: e.response,
        type: e.type,
        error: e.error,
      );

  @override
  String toString() {
    return 'Receive timeout in connection with API server, please try again';
  }
}

/// DioErrorType.response && statusCode == 400
class BadRequestError extends MorphNetworkError {
  BadRequestError({
    required super.requestOptions,
    super.response,
    super.type,
    super.error,
  });

  factory BadRequestError.fromDioError(DioError e) => BadRequestError(
        requestOptions: e.requestOptions,
        response: e.response,
        type: e.type,
        error: e.error,
      );

  @override
  String toString() {
    return 'Bad request';
  }
}

/// DioErrorType.response && (statusCode == 401 || statusCode == 403)
class UnauthorizedError extends MorphNetworkError {
  UnauthorizedError({
    required super.requestOptions,
    super.response,
    super.type,
    super.error,
  });

  factory UnauthorizedError.fromDioError(DioError e) => UnauthorizedError(
        requestOptions: e.requestOptions,
        response: e.response,
        type: e.type,
        error: e.error,
      );

  @override
  String toString() {
    return 'Access denied';
  }
}

/// DioErrorType.response && statusCode == 404
class NotFoundError extends MorphNetworkError {
  NotFoundError({
    required super.requestOptions,
    super.response,
    super.type,
    super.error,
  });

  factory NotFoundError.fromDioError(DioError e) => NotFoundError(
        requestOptions: e.requestOptions,
        response: e.response,
        type: e.type,
        error: e.error,
      );

  @override
  String toString() {
    return 'The requested information could not be found';
  }
}

/// DioErrorType.response && statusCode == 409
class ConflictError extends MorphNetworkError {
  ConflictError({
    required super.requestOptions,
    super.response,
    super.type,
    super.error,
  });

  factory ConflictError.fromDioError(DioError e) => ConflictError(
        requestOptions: e.requestOptions,
        response: e.response,
        type: e.type,
        error: e.error,
      );

  @override
  String toString() {
    return 'Conflict occurred';
  }
}

/// DioErrorType.response && statusCode == 500
class InternalServerErrorError extends MorphNetworkError {
  InternalServerErrorError({
    required super.requestOptions,
    super.response,
    super.type,
    super.error,
  });

  factory InternalServerErrorError.fromDioError(DioError e) =>
      InternalServerErrorError(
        requestOptions: e.requestOptions,
        response: e.response,
        type: e.type,
        error: e.error,
      );

  @override
  String toString() {
    return 'Unknown error occurred, please try again later';
  }
}

/// DioErrorType.response && statusCode == 503
class ServiceUnavailableError extends MorphNetworkError {
  ServiceUnavailableError({
    required super.requestOptions,
    super.response,
    super.type,
    super.error,
  });

  factory ServiceUnavailableError.fromDioError(DioError e) =>
      ServiceUnavailableError(
        requestOptions: e.requestOptions,
        response: e.response,
        type: e.type,
        error: e.error,
      );

  @override
  String toString() {
    return 'Service unavailable';
  }
}

/// The others
class UnexpectedError extends MorphNetworkError {
  UnexpectedError({
    required super.requestOptions,
    super.response,
    super.type,
    super.error,
  });

  factory UnexpectedError.fromDioError(DioError e) => UnexpectedError(
        requestOptions: e.requestOptions,
        response: e.response,
        type: e.type,
        error: e.error,
      );

  @override
  String toString() {
    return 'Unexpected error';
  }
}

/// error response
class MorphNetworkResponseError extends MorphNetworkError {
  final MorphNetworkErrorType errorType;
  final int subCode;
  final String errorMessage;
  final dynamic detail;

  MorphNetworkResponseError({
    required super.requestOptions,
    required this.errorType,
    required this.subCode,
    required this.errorMessage,
    this.detail,
  });

  @override
  String toString() {
    return jsonEncode(<String, dynamic>{
      'errorType': errorType.name,
      'subCode': subCode,
      'message': errorMessage,
      'detail': detail,
    });
  }

  factory MorphNetworkResponseError.fromResponseBody(
    RequestOptions requestOptions,
    Map<String, dynamic> responseBody,
  ) {
    final errorType = MorphNetworkErrorType.values
        .byName(ReCase(responseBody['error'] as String).camelCase);
    final subCode = responseBody['subCode'] as int;
    final errorMessage = responseBody['message'] as String;
    final detail = responseBody['detail'];
    return MorphNetworkResponseError(
      requestOptions: requestOptions,
      errorType: errorType,
      subCode: subCode,
      errorMessage: errorMessage,
      detail: detail,
    );
  }
}

/// User not found
class BadRequestDetailResponseError extends MorphNetworkResponseError {
  BadRequestDetailResponseError({
    required super.requestOptions,
    required super.errorType,
    required super.subCode,
    required super.errorMessage,
    super.detail,
  });

  factory BadRequestDetailResponseError.fromResponseError(
    MorphNetworkResponseError responseError,
  ) =>
      BadRequestDetailResponseError(
        requestOptions: responseError.requestOptions,
        errorType: responseError.errorType,
        subCode: responseError.subCode,
        errorMessage: responseError.errorMessage,
        detail: responseError.detail,
      );
}

/// User not found
class UserpoolUserNotFoundResponseError extends MorphNetworkResponseError {
  UserpoolUserNotFoundResponseError({
    required super.requestOptions,
    required super.errorType,
    required super.subCode,
    required super.errorMessage,
    super.detail,
  });

  factory UserpoolUserNotFoundResponseError.fromResponseError(
    MorphNetworkResponseError responseError,
  ) =>
      UserpoolUserNotFoundResponseError(
        requestOptions: responseError.requestOptions,
        errorType: responseError.errorType,
        subCode: responseError.subCode,
        errorMessage: responseError.errorMessage,
        detail: responseError.detail,
      );
}
