import 'package:dio/dio.dart';
import 'package:morph_dart/src/enums/record.dart';
import 'package:morph_dart/src/models/record/filter_condition.dart';
import 'package:morph_dart/src/models/record/record.dart';

class RecordMethods<Fields> {
  final Dio dio;
  final String tableSlug;
  final Object Function(Fields) toJsonFields;
  final Fields Function(Object?) fromJsonFields;

  const RecordMethods(
    this.dio,
    this.tableSlug,
    this.toJsonFields,
    this.fromJsonFields,
  );

  Future<RecordListResponse<Fields>> queryRecords({
    required QueryRecordRequestBody body,
  }) async {
    final response = await dio
        .post<Map<String, dynamic>>(
          '/record/query',
          options: Options(
            headers: {
              'tableSlug': tableSlug,
            },
          ),
          data: body,
        )
        .then(
          (res) => RecordListResponse<Fields>.fromJson(
            res.data!,
            fromJsonFields,
          ),
        );

    return response;
  }

  Future<RecordResponse<Fields>?> findRecord({
    required String id,
  }) async {
    final response = await dio
        .post<Map<String, dynamic>>(
          '/record/query',
          options: Options(
            headers: {
              'tableSlug': tableSlug,
            },
          ),
          data: QueryRecordRequestBody(
            filter: FilterCondition.and(
              [
                FilterConditionUnit.text(
                  property: 'id',
                  operator: TextFilterOperator.equal,
                  value: id,
                )
              ],
            ),
          ),
        )
        .then(
          (res) => RecordListResponse<Fields>.fromJson(
            res.data!,
            fromJsonFields,
          ),
        );
    if (response.records.isEmpty) {
      return null;
    } else {
      return RecordResponse(
        id: response.records[0].id,
        fields: response.records[0].fields,
        createdAt: response.records[0].createdAt,
        updateAt: response.records[0].updateAt,
        createdBy: response.records[0].createdBy,
      );
    }
  }

  Future<DownloadRecordResponse> downloadRecords({
    required DownloadRecordRequestBody body,
  }) async {
    final response = await dio
        .post<Map<String, dynamic>>(
          '/record/download',
          options: Options(
            headers: {
              'tableSlug': tableSlug,
            },
          ),
          data: body,
        )
        .then((res) => DownloadRecordResponse.fromJson(res.data!));

    return response;
  }

  Future<RecordResponse<Fields>> createRecord({
    required CreateRecordRequestBody<Fields> body,
  }) async {
    final response = await dio
        .post<Map<String, dynamic>>(
          '/record',
          options: Options(
            headers: {
              'tableSlug': tableSlug,
            },
          ),
          data: body.toJson(toJsonFields),
        )
        .then(
          (res) => RecordResponse<Fields>.fromJson(
            res.data!,
            fromJsonFields,
          ),
        );

    return response;
  }

  Future<RecordResponse<Fields>> updateRecord({
    required String recordId,
    required UpdateRecordRequestBody<Fields> body,
  }) async {
    final response = await dio
        .put<Map<String, dynamic>>(
          '/record/$recordId',
          options: Options(
            headers: {
              'tableSlug': tableSlug,
            },
          ),
          data: body.toJson(toJsonFields),
        )
        .then(
          (res) => RecordResponse.fromJson(
            res.data!,
            fromJsonFields,
          ),
        );

    return response;
  }

  Future<bool> deleteRecord({
    required String recordId,
  }) async {
    final response = await dio
        .delete<Map<String, dynamic>>(
          '/record/$recordId',
          options: Options(
            headers: {
              'tableSlug': tableSlug,
            },
          ),
        )
        .then((res) => res.data!['message'] == 'ok');

    return response;
  }
}
