import 'dart:io';

import 'package:dio/dio.dart';
import 'package:morph_dart/src/error/network_error.dart';
import 'package:morph_dart/src/network/auth_token_config.dart';

class BaseInterceptor extends QueuedInterceptor {
  BaseInterceptor(this.interceptors, this.tokenConfig);

  final List<Interceptor> interceptors;
  final AuthTokenConfig tokenConfig;

  @override
  Future<void> onRequest(
    RequestOptions options,
    RequestInterceptorHandler handler,
  ) async {
    try {
      final idToken = await tokenConfig.getIdToken();
      options.headers[HttpHeaders.authorizationHeader] = 'Bearer $idToken';
    } on Exception catch (_) {
      return handler.reject(UnauthorizedError(requestOptions: options));
    }
    return handler.next(options);
  }

  @override
  Future<void> onError(DioError err, ErrorInterceptorHandler handler) async {
    final requestOptions = err.requestOptions;
    final errResponse = err.response;
    if (err.type == DioErrorType.response && errResponse != null) {
      if (errResponse.statusCode == 401 || errResponse.statusCode == 403) {
        late final String? idToken;
        try {
          await tokenConfig.refresh();
          idToken = await tokenConfig.getIdToken();
        } on Exception catch (_) {
          return handler.next(UnauthorizedError.fromDioError(err));
        }
        requestOptions.headers[HttpHeaders.authorizationHeader] =
            'Bearer $idToken';
        try {
          final retryDio = Dio(
            BaseOptions(
              baseUrl: requestOptions.baseUrl,
              method: requestOptions.method,
              contentType: requestOptions.contentType,
              extra: requestOptions.extra,
              followRedirects: requestOptions.followRedirects,
              headers: requestOptions.headers,
              listFormat: requestOptions.listFormat,
              maxRedirects: requestOptions.maxRedirects,
              receiveDataWhenStatusError:
                  requestOptions.receiveDataWhenStatusError,
              receiveTimeout: requestOptions.receiveTimeout,
              requestEncoder: requestOptions.requestEncoder,
              responseDecoder: requestOptions.responseDecoder,
              responseType: requestOptions.responseType,
              sendTimeout: requestOptions.sendTimeout,
              connectTimeout: requestOptions.connectTimeout,
              validateStatus: requestOptions.validateStatus,
            ),
          )..interceptors.addAll(interceptors);

          final response = await retryDio.request<dynamic>(
            requestOptions.path,
            cancelToken: requestOptions.cancelToken,
            onReceiveProgress: requestOptions.onReceiveProgress,
            data: requestOptions.data,
            queryParameters: requestOptions.queryParameters,
          );
          return handler.resolve(response);
        } on DioError catch (e) {
          return handler.next(e);
        }
      } else {
        return handler.next(err);
      }
    }
    return super.onError(err, handler);
  }
}
